#include "word_list.h"

#include <algorithm>
#include <fstream>
#include <random>

WordList &WordList::multiply(size_t factor) {
  WordList result;
  result.reserve(size() * factor);

  for (auto time = 0; time < factor; ++time) {
    std::copy(cbegin(), cend(), std::back_inserter(result));
  }

  std::swap(*this, result);

  return *this;
};

WordList &WordList::shuffle() {
  std::random_device random_device;
  std::mt19937 random_number_generator(random_device());
  std::shuffle(begin(), end(), random_number_generator);

  return *this;
}

WordList WordList::fourCaps() {
  const static std::string charset_ = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";

  WordList word_list;
  word_list.reserve(std::pow(charset_.length(), 4));

  for (auto char_1 : charset_) {
    for (auto char_2 : charset_) {
      for (auto char_3 : charset_) {
        for (auto char_4 : charset_) {
          word_list.emplace_back(
              std::initializer_list<char>({char_1, char_2, char_3, char_4}));
        }
      }
    }
  }

  return word_list;
};

WordList WordList::fromFile(const std::filesystem::path &path) {
  WordList word_list;

  std::ifstream ifs(path);

  std::string line;
  while (std::getline(ifs, line)) {
    word_list.emplace_back(line);
  }

  return word_list;
}

void WordList::find_prefix_in_range(
    const WordList &word_list, const std::string_view &search_prefix,
    size_t start_index, size_t end_index,
    std::forward_list<const std::string *> &result, std::mutex &result_mutex) {
  std::forward_list<const std::string *> local_results;

  for (size_t index = start_index; index < end_index; ++index) {
    const auto &current_word = word_list[index];
    if (current_word.starts_with(search_prefix)) {
      local_results.push_front(&current_word);
    }
  }

  if (!local_results.empty()) {
    const std::lock_guard<std::mutex> lock(result_mutex);
    result.merge(local_results);
  }
};

void WordRefList::find_prefix_in_range(
    const WordRefList &word_list, const std::string_view &search_prefix,
    size_t start_index, size_t end_index,
    std::forward_list<const std::string *> &result, std::mutex &result_mutex) {
  std::forward_list<const std::string *> local_results;

  for (size_t index = start_index; index < end_index; ++index) {
    const auto *current_word = word_list[index];
    if (current_word->starts_with(search_prefix)) {
      local_results.push_front(current_word);
    }
  }

  if (!local_results.empty()) {
    const std::lock_guard<std::mutex> lock(result_mutex);
    result.merge(local_results);
  }
};
